#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

#   Copyright 2013 Rodrigo J. da Fonseca
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__version__ = '2.0b5'

from os.path import basename
from urlparse import urlsplit
from urlparse import urlparse
from pdfminer.pdfinterp import PDFResourceManager, process_pdf
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from cStringIO import StringIO
from os.path import expanduser, join
from HTMLParser import HTMLParser
from tempfile import TemporaryFile
import re
import os
import time
import locale
import codecs
import json
import urllib2
import sys

# Add new units here
sesc_unit = {
    'Pinheiros' : {
        'url'     : 'http://www.sescsp.org.br/unidades/unidade-facilidades.action?id=10',
        'pdf'     : r'(?i).*?card.+?pio.*?',
        'entries' : r'^.?\d\d\s•.+?(?=\n\n)',
        'day'     : r'(?P<day>\d\d?)\s',
        'menus'   : r'brasileirinho\s*?\n((?:.|\n)+?)\nfog.*?o\scultural\s*\n((?:.|\n)+)'
    },
}

# You may edit the default values if you like
default_unit = 'Pinheiros'
default_home = join(expanduser("~"),'.sesc')
cache_filename = 'cache.json'

# But do not edit below this point, if you don't know what you are doing

locale.setlocale(locale.LC_ALL, '')
cache_version = 3
context = {'version' : cache_version}
quiet = True

# Wrap sys.stdout into a StreamWriter to allow writing unicode.
# Thanks to EOL: http://stackoverflow.com/questions/4545661/unicodedecodeerror-when-redirecting-to-file
sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)

class MyHTMLParser(HTMLParser):
    def __init__ (self, pattern):
        HTMLParser.__init__(self)
        self.inside_a = False
        self.extract = False
        self.urls = []
        self.href = None
        self.match = re.compile(pattern)

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            self.inside_a = True
            for (name, value) in attrs:
                if name == 'href':
                    self.href = value

    def handle_endtag(self, tag):
        if tag == 'a':
            if self.inside_a:
                if self.extract and self.href:
                    self.urls.append(self.href)
                self.inside_a = False
                self.extract = False
                self.href = None

    def handle_data(self, data):
        if self.inside_a:
            if self.match.match(data):
                self.extract = True

    def get_urls(self, text):
        self.feed(text)
        return self.urls

def log(message):
    global quiet
    if not quiet:
        if not isinstance(message, unicode):
            message = unicode(message, 'utf-8')
        print(message)

def write(message):
    if not isinstance(message, unicode):
        message = unicode(message, 'utf-8')
    print(message)

# http://stackoverflow.com/questions/2647723/urllib2-to-string
def get_pdf_address(url, regex):
    """Extract the PDF's download link from the web page.

    Solution extracted from: http://stackoverflow.com/questions/2647723/urllib2-to-string
    
    Keyword arguments:
    url   -- String of the URL for the web page from where to extract the download link
    regex -- String of a regular expression to match the downlaod tag

    """
    f = urllib2.urlopen(url)
    s = f.read()
    parser = MyHTMLParser(regex)
    urls = parser.get_urls(s)
    if len(urls) == 0:
        raise Exception('PDF URL not found')
    pdf_url = urls[-1]
    if pdf_url[0] == '/':
        ps = urlparse(url)
        pdf_url = ps[0]+'://'+ps[1]+pdf_url
    return pdf_url

def get_current_day():
    return time.strftime('%d')

def get_pdf_as_text(url):
    """ Convert a PDF into a string."""

    with TemporaryFile() as fp:

        req = urllib2.Request(url)
        r = urllib2.urlopen(req)

        fp.write(r.read())
        fp.flush()

        rsrcmgr = PDFResourceManager()
        retstr = StringIO()
        device = TextConverter(rsrcmgr, retstr, codec='utf-8', laparams=LAParams())

        process_pdf(rsrcmgr, device, fp)
        device.close()

        text = retstr.getvalue()
        retstr.close()

    return text

def save(context_file):
    with open(context_file, 'w') as f:
        json.dump(context, f)

def load(context_file):
    if os.path.exists(context_file):
        with open(context_file, 'r') as f:
            global context
            context = json.load(f)

def fetch(unit):
    log('Verificando se há uma atualização do cardápio...')
    unit_data = context.get(unit) or {}
    pdf_url = get_pdf_address(sesc_unit[unit]['url'], sesc_unit[unit]['pdf'])
    if pdf_url == unit_data.get('pdf'):
        log('Você já possui o cardápio mais recente.')
        return
    else:
        log('Baixando novo cardápio...')
    content = get_pdf_as_text(pdf_url)
    log('Novo cardápio baixado com sucesso.')
    data, date_from, date_to = extract_data(content, sesc_unit[unit])
    unit_data['data'] = data
    unit_data['pdf'] = pdf_url
    unit_data['from'] = date_from
    unit_data['to'] = date_to
    context[unit] = unit_data
    return

def extract_data(text, unit):
    log('Extraindo dados do cardápio...')
   
    # Initial setup
    data = {}
    date_from = None
    date_to = None
    
    # Extracting extries
    entries = re.findall(unit['entries'], text, re.MULTILINE | re.DOTALL)
    for entry in entries:
        day = re.search(unit['day'], entry.strip() , re.IGNORECASE | re.DOTALL).group('day').replace('\n','')
        menus =  [menu.replace('\n','') for menu in re.search(unit['menus'], entry.strip(), re.IGNORECASE | re.DOTALL).groups()]
        value = '\n'.join(menus)

        key = day
        data[key] = value

        if date_from is None or key < date_from:
            date_from = key
        if date_to is None or key > date_to:
            date_to = key
    return data, date_from, date_to

def show_dialog(title, unit, date):
    from Tkinter import W, E, BOTH, SUNKEN, StringVar
    from ttk import Frame, Label, Combobox
    from tkFont import Font

    class MenuDisplay(Frame):
        def __init__(self, title, unit, date, master=None):
            Frame.__init__(self, master)
            
            if not master:
                self.master.title(title)
                self.master.maxsize(400, 150)
                self.master.minsize(400, 150)
            
            self.key = StringVar()
            self.menu = StringVar()
            self.unit = StringVar()
            
            self.unit_combo = Combobox(self, textvariable=self.unit, state='readonly', font=Font(weight='bold'))
            self.unit_combo['values'] = [u for u in context.keys() if u != 'version']
            self.unit_combo.bind('<<ComboboxSelected>>', self.update_unit)
            self.unit_combo.grid(row=0, column=0, sticky=W+E, padx=10, pady=4) 

            self.date = Combobox(self, textvariable=self.key, state='readonly', font=Font(weight='bold'))
            
            self.date.bind('<<ComboboxSelected>>', self.update_menu)
            self.date.grid(row=1, column=0, sticky=W+E, padx=10, pady=4) 
            
            self.data_label = Label(self, textvariable=self.menu, wraplength=self.master.maxsize()[0] - 10)
            self.data_label.grid(row=2, column=0, sticky=W+E, padx=10, pady=0)
            
            self.pack(fill=BOTH, expand=1)
            
            try:
                self.unit.set(unit)
                self.update_unit()
                self.date.current(self.date['values'].index(date))
                self.update_menu()
            except ValueError:
                self.key.set(date)
                self.menu.set('Não há menu disponível para esse dia.')
            
            if not master:
                self.mainloop()

        def update_menu(self, *args):
            self.menu.set(context[self.unit.get()]['data'].get(self.key.get()))

        def update_unit(self, *args):
            self.date['values'] = [key in sorted(context[self.unit.get()]['data'].iterkeys())]
            self.update_menu()
    
    MenuDisplay(title, unit, date)

def notify_all(guser, gpass, unit, day, menu, email_list):
    import smtplib
    from email.mime.text import MIMEText
    from email.header import Header
    msg = '%s\n\n--------------------------------\nMensagem gerada automaticamente.\n--------------------------------\n' % menu
    msg = MIMEText(msg, 'plain', 'utf-8')
    msg['From'] = guser
    msg['Subject'] = Header('Cardápio do SESC %s para o dia %s' % (unit, day), 'utf-8')

    smtpserver = smtplib.SMTP("smtp.gmail.com",587)
    smtpserver.starttls()
    smtpserver.login(guser, gpass)
    smtpserver.sendmail(guser, email_list, msg.as_string())

def main(argv):
    import getopt
    def usage():
        write('usage: %s [options]' % argv[0])
        write('')
        write('Exibe o cardápio, do SESC de Pinheiros, para o dia corrente.')
        write('')
        write('Opções:')
        write('-a, --all              Exibe o cardápio para todos os dias.')
        write('-c, --clean            Limpa o cache antes de buscar o cardápio.')
        write('-d, --day=day          Exibe cardápio para um dia específico ')
        write('-g, --gui              Exibe o cardápio em uma janela. Não é compatível com a opção "-a".')
        write('    --guser=username   Configura o nome de usuário do gmail para o envio de e-mail.')
        write('    --gpass=senha      Configura a senha de aplicativo do gmail para o envio de e-mail.')
        write('-h, --help             Exibe esta ajuda.')
        write('-n, --notify=lista     Lista de e-mails, separados por ponto e vírgula, a serem notificados do menu.')
        write('-p, --path=caminho     Caminho onde armazenar os arquivos, padrão "%s".' % default_home)
        write('-q, --quiet            Modo silencioso. Não exibe mensagens intermediárias.')
        write('-u, --unit=unidade     Unidade do SESC, padrão unidade de "%s". Unidades cadastradas: %s' % (default_unit , ', '.join(sesc_unit.keys())))
        write('-v, --version          Exibe a versão do programa.')
        return 100
    try:
        (opts, args) = getopt.getopt(argv[1:], 'acd:ghn:p:qu:v',['all', 'clean', 'day=', 'format=', 'gui', 'guser=', 'gpass=', 'help', 'notify=', 'path=', 'quiet', 'unit=', 'version'])
    except getopt.GetoptError:
        return usage()

    def version():
        write('Versão do programa: %s' % __version__)
        write('Versão do cache: %s' % cache_version)
        return 100

    global context, quiet, date_format
    show_all = False
    unit = default_unit
    home = default_home
    clean = False
    quiet = False
    key = get_current_day()
    gui = False
    notify = None
    guser = None
    gpass = None

    for (k, v) in opts:
        if k == '-a' or k == '--all': show_all = True
        elif k == '-c' or k == '--clean': clean = True
        elif k == '-d' or k == '--date': key = v
        elif k == '-g' or k == '--gui': gui = True
        elif k == '--guser': guser = v
        elif k == '--gpass': gpass = v
        elif k == '-h' or k == '--help': return usage()
        elif k == '-n' or k == '--notify': notify = v.strip(';').split(';')
        elif k == '-p' or k == '--path': home = v
        elif k == '-q' or k == '--quiet': quiet = True
        elif k == '-u' or k == '--unit': unit = v
        elif k == '-v' or k == '--version': return version()

    if unit not in sesc_unit.keys():
        write('Unidade inválida: %s' % unit)
        write('Experimente uma das seguintes unidades: %s' % ', '.join(sesc_unit.keys()))
        sys.exit(0)

    if not os.path.exists(home):
        log('Criando diretório "%s".' % home)
        os.makedirs(home)

    context_file = join(home, cache_filename)

    if clean:
        log('Descartando cache local...')
        if os.path.exists(context_file):
            os.remove(context_file)

    load(context_file)

    if cache_version != context.get('version'):
        log('Versão do cache inconsistente. Limpando cache...')
        if os.path.exists(context_file):
            os.remove(context_file)
        context = {'version' : cache_version}

    if unit in context and key < context[unit]['from']:
        write('')
        write(key)
        write('Data antiga demais. Não há cardápio disponível.')
        write('')
        sys.exit(1)

    if unit not in context or key > context[unit]['to']:
        log('O menu para o dia não foi encontrado no cache local.')
        fetch(unit)
        save(context_file)

    if show_all:
        write('')
        for key in sorted(context[unit]['data'].iterkeys()):
            write(key)
            write(context[unit]['data'][key])
            write('')
    else:

        if notify and guser and gpass:
            data = context[unit]['data'].get(key) or 'Não há menu disponível para esse dia.'
            write('Notificando as seguintes pessoas: %s' % notify)
            notify_all(guser, gpass, unit,  key, data, notify)

        if gui:
            show_dialog(os.path.basename(argv[0]), unit, key)
        else:
            write('')
            write(key)
            data = context[unit]['data'].get(key) or 'Não há menu disponível para esse dia.'
            write(data)
            write('')
        
        
    return 0

if __name__ == '__main__': sys.exit(main(sys.argv))
